/*
Build under Linux: gcc -o pi pi.c
Fire up:           ./pi
*/

#include <stdio.h>
#include <stdlib.h>

#define M_PI 3.14159265358979323846264338327			//from The GNU C Library Reference Manual, p. 467

short int init_values_32bit(long double *x,double *y,float *z);

int main(int argc, char **argv)
{
  //let the compiler do the work

  long double xld_gcc = M_PI;
  double xd_gcc = M_PI;
  float xf_gcc = M_PI;

  //let the FPU do the work

  long double xld_fpu;
  double xd_fpu;
  float xf_fpu;
  short int cwfpu;
  cwfpu = init_values_32bit(&xld_fpu,&xd_fpu,&xf_fpu);

  //print results

  printf("\nLayout of data types:\n");
  printf("----------------------\n");
  printf("Long Double (REAL 10)   = %i Bytes\n",sizeof(long double));
  printf("Double (REAL 8)         = %i Bytes\n",sizeof(double));
  printf("Float (REAL 4)          = %i Bytes\n\n",sizeof(float));
  printf("Compiler Results:\n");
  printf("-----------------\n");
  printf("PI                      = 3.14159265358979323846264338327 ...\n");
  printf("PI as Long Double Value = %1.19Lf\n",xld_gcc);
  printf("PI as Double Value      = %1.16lf\n",xd_gcc);
  printf("PI as Float Value       = %1.6f\n\n",xf_gcc);

  printf("FPU Results:\n");
  printf("------------\n");
  printf("FPU control word:       = %X hex\n",cwfpu);
  printf("PI                      = 3.14159265358979323846264338327 ...\n");
  printf("PI as Long Double Value = %1.19Lf\n",xld_fpu);
  printf("PI as Double Value      = %1.16lf\n",xd_fpu);
  printf("PI as Float Value       = %1.6f\n\n",xf_fpu);

  return 0;
}

/*
init_values_32bit
Purpose:	Initialize floating point values via FPU.
Input:		[ebp+8]  = pointer to long double
		[ebp+12] = pointer to double
		[ebp+16] = pointer to float
*/
short int init_values_32bit(long double *x,double *y,float *z)
{
  short int cw;							//FPU control word
  __asm__ __volatile__(
    ".intel_syntax noprefix 			\n\t"
    "fstcw	[ebp-2]				\n\t"		//store current control word
    "mov	eax,[ebp+8] 			\n\t"		//eax -> long double
    "mov	ecx,[ebp+12] 			\n\t"		//ecx -> double
    "mov	edx,[ebp+16] 			\n\t"		//edx -> float
    "fldpi 					\n\t"		//st0 = PI
    "fst	dword ptr [edx] 		\n\t"		//store float
    "fst	qword ptr [ecx] 		\n\t"		//store double
    "fstp	tbyte ptr [eax]			\n\t"		//store long double
    ".att_syntax prefix 			\n\t"
    :
    :"g"(x),"g"(y),"g"(z)
    :"%eax","%ecx","%edx"
    );
  return cw;
}

